//define encoder 0 variables
//encoder 0 is motor encoder. original 1024CPR to 4096CPR with interrupts
#define encoder0PinA 3
#define encoder0PinB 2
volatile int encoder0Count = 0;
volatile float angle0 = 0;
volatile float velocity0 = 0;
volatile float angle0_previous = 0;
volatile float angle0_post = 0;

//define encoder 1 variables
//encoder 1 is pendulum encoder. original 1024CPR to 4096CPR with interrupts
#define encoder1PinA 20
#define encoder1PinB 21
volatile int encoder1Count = 0;
volatile float angle1 = 0;
volatile float velocity1 = 0;
volatile float angle1_previous = 0;
volatile float angle1_post = 0;

//define timer2 variables and timer counter
volatile int tcnt2 = 131;  //CPU frequency/prescaler= 8us >>> desired period/8us=125
volatile int t = 0;  //MAX(uint8)+1-125=131

//define system input variables
//k's are feedback gains
//result from LQR, when Q = diag([1.5 8 0 0]); R = 0.0028;
volatile float proportional = 0;
volatile float derivative = 0;
volatile float integral = 0;
float kp = 50;
float kd = 8;
float ki = 40;
float u = 0;
float motor_input = 0;
int INPUT_MAX = 3200;

void setup()
{
  //motor control code
  
  // initialize software serial object with baud rate of 38.4 kbps
  Serial1.begin(38400);
  // the simple motor controller must be running for at least 1 ms
  // before we try to send serial data, so we delay here for 5 ms
  delay(5);
  // motor controller automatic baud detection
  // send it the byte 170 in decimal so that it can learn the baud rate
  Serial1.write(0xAA);
  // exit safe start command clears the safe-start violation and lets the motor run
  
  exitSafeStart();

  //for encoder 0 set inputs and interrupts
  pinMode(encoder0PinA, INPUT); 
  pinMode(encoder0PinB, INPUT); 
  attachInterrupt(0, doEncoder0A, CHANGE);  
  attachInterrupt(1, doEncoder0B, CHANGE);  
  //for encoder 1 set inputs and interrupts
  pinMode(encoder1PinA, INPUT); 
  pinMode(encoder1PinB, INPUT); 
  attachInterrupt(2, doEncoder1A, CHANGE);  
  attachInterrupt(3, doEncoder1B, CHANGE);    
  
  //timer 2 from atmega 2560 data sheet
  //interrupts every 1 ms
  TIMSK2 &= ~(1<<TOIE2);
  TCCR2A &= ~((1<<WGM21) | (1<<WGM20));
  TCCR2B &= ~(1<<WGM22);
  ASSR &= ~(1<<AS2);
  TIMSK2 &= ~(1<<OCIE2A);
  TCCR2B |= (1<<CS22)  | (1<<CS20);
  TCCR2B &= ~(1<<CS21);
  TCNT2 = tcnt2;
  TIMSK2 |= (1<<TOIE2);

  //for debugging in serial monitor outputs error's and motor input
  Serial.begin (9600); 
}

void loop()
{
  u = - (kp*proportional + kd*derivative + ki*integral);
  motor_input = 3200 / 19.5 * u; // because a 19.5V power supply is used to power the motor at max 3200 rate and u is motor input
  
  // 3200 : 19.5V = motor_input : u, thus motor_input = 5600*u/19.5
  motor_input = (int)motor_input;
  
  if (motor_input >= INPUT_MAX){
    motor_input = INPUT_MAX;
  }
  else if (motor_input <= -INPUT_MAX){
    motor_input = -INPUT_MAX;
  }
  
  
  Serial.print(proportional);
  Serial.print(" ");
  Serial.print(derivative);
  Serial.print(" ");
  Serial.print(integral);
  Serial.print(" ");
  Serial.println(motor_input);
  
  
  setMotorSpeed(motor_input); // full speed is 3200
}
